@@ -1,19 +1,16 @@
 (defpackage "SB-POSIX-TESTS"
-  (:use "COMMON-LISP" "SB-RT"))
+  (:import-from #:test-util #:deftest)
+  (:use "COMMON-LISP"))
 
 (in-package "SB-POSIX-TESTS")
 
-(defvar *test-directory*
-  (ensure-directories-exist
-   (merge-pathnames (make-pathname :directory '(:relative "test-lab"))
-                    (make-pathname :directory
-                                   (pathname-directory *load-truename*)))))
+(defvar *test-directory* test-util:*test-directory*)
 
 (defvar *current-directory* *default-pathname-defaults*)
 
 (defvar *this-file* *load-truename*)
 
-(eval-when (:compile-toplevel :load-toplevel)
+(eval-when (:compile-toplevel :load-toplevel :execute)
   (defconstant +mode-rwx-all+
     (logior sb-posix::s-irusr sb-posix::s-iwusr sb-posix::s-ixusr
             #-win32
@@ -22,6 +19,7 @@
              sb-posix::s-iroth sb-posix::s-iwoth sb-posix::s-ixoth))))
 
 (defmacro define-eacces-test (name form &rest values)
+  #+win32 (declare (ignore name form values))
   #-win32
   `(deftest ,name
     (block ,name
@@ -107,13 +105,16 @@
   (handler-case
       (sb-posix:mkdir #-win32 "/" #+win32 "C:/" 0)
     (sb-posix:syscall-error (c)
-      (sb-posix:syscall-errno c)))
-  #+darwin
-  #.sb-posix:eisdir
-  #+win32
-  #.sb-posix:eacces
-  #-(or darwin win32)
-  #.sb-posix::eexist)
+      ;; The results aren't the most consistent ones across platforms. Darwin
+      ;; likes EISDIR, Windows either EACCESS or EEXIST, others EEXIST.
+      ;; ...let's just accept them all.
+      (when (member (sb-posix:syscall-errno c)
+                    (list #.sb-posix:eisdir
+                          #.sb-posix:eacces
+                          #.sb-posix::eexist)
+                    :test #'eql)
+        :ok)))
+  :ok)
 
 (define-eacces-test mkdir.error.3
   (let* ((dir (merge-pathnames
@@ -154,29 +155,25 @@
         (sb-posix:syscall-errno c))))
   #.sb-posix::enoent)
 
-(deftest rmdir.error.2
-  (handler-case
-      (sb-posix:rmdir *this-file*)
-    (sb-posix:syscall-error (c)
-      (sb-posix:syscall-errno c)))
-  #-win32
-  #.sb-posix::enotdir
-  #+win32
-  #.sb-posix::einval)
+(test-util:with-test (:name :rmdir.error.2)
+  (let ((err (handler-case (sb-posix:rmdir *this-file*)
+               (sb-posix:syscall-error (c) (sb-posix:syscall-errno c)))))
+    (assert #-unix (eql err #.sb-posix:einval)
+            ;; non-writable parent dir may return EACCES instead of ENOTDIR
+            #+unix (member err `(,sb-posix:enotdir ,sb-posix:eacces)))))
 
+#-haiku
 (deftest rmdir.error.3
   (handler-case
-      (sb-posix:rmdir #-win32 "/" #+win32 "C:/")
+      (sb-posix:rmdir #-win32 "/" #+win32 (sb-ext:posix-getenv "windir"))
     (sb-posix:syscall-error (c)
-      (sb-posix:syscall-errno c)))
-  #+darwin
-  #.sb-posix:eisdir
-  #+win32
-  #.sb-posix::eacces
-  #+sunos
-  #.sb-posix::einval
-  #-(or darwin win32 sunos)
-  #.sb-posix::ebusy)
+      (typep
+       (sb-posix:syscall-errno c)
+       `(member #+(or darwin openbsd freebsd) ,sb-posix:eisdir
+                #+win32 ,sb-posix::eacces #+win32 ,sb-posix::enotempty
+                #+sunos ,sb-posix::einval
+                #-(or darwin openbsd freebsd win32 sunos haiku) ,sb-posix::ebusy))))
+  t)
 
 (deftest rmdir.error.4
   (let* ((dir (ensure-directories-exist
@@ -219,6 +216,7 @@
         result)))
   #.sb-posix::eacces)
 
+#-(or (and darwin x86) win32)
 (deftest stat.1
   (let* ((stat (sb-posix:stat *test-directory*))
          (mode (sb-posix::stat-mode stat)))
@@ -226,15 +224,12 @@
     (logand mode (logior sb-posix::s-iread sb-posix::s-iwrite sb-posix::s-iexec)))
   #.(logior sb-posix::s-iread sb-posix::s-iwrite sb-posix::s-iexec))
 
-#-win32
+#-(or (and darwin x86) win32 haiku)
 (deftest stat.2
-  (let* ((stat (sb-posix:stat "/"))
-         (mode (sb-posix::stat-mode stat)))
-    ;; it's logically possible for / to be writeable by others... but
-    ;; if it is, either someone is playing with strange security
-    ;; modules or they want to know about it anyway.
-    (logand mode sb-posix::s-iwoth))
-  0)
+  (eql
+   (sb-posix::stat-mode (sb-posix:stat "/"))
+   (sb-posix::stat-mode (sb-posix:stat (make-pathname :directory '(:absolute :up)))))
+  t)
 
 (deftest stat.3
   (let* ((now (get-universal-time))
@@ -247,16 +242,6 @@
     (< (- atime unix-now) 10))
   t)
 
-#-win32
-(deftest stat.4
-  (let* ((stat (sb-posix:stat (make-pathname :directory '(:absolute :up))))
-         (mode (sb-posix::stat-mode stat)))
-    ;; it's logically possible for / to be writeable by others... but
-    ;; if it is, either someone is playing with strange security
-    ;; modules or they want to know about it anyway.
-    (logand mode sb-posix::s-iwoth))
-  0)
-
 ;; Test that stat can take a second argument.
 #-win32
 (deftest stat.5
@@ -273,14 +258,15 @@
 
 #+win32
 (deftest stat.5
-    (let* ((stat-1 (sb-posix:stat "/"))
-           (mode-1 (sb-posix:stat-mode stat-1))
-           (stat-2 (sb-posix:stat "C:\\CONFIG.SYS"
-                                   stat-1))
-           (mode-2 (sb-posix:stat-mode stat-2)))
-      (values
-       (eq stat-1 stat-2)
-       (/= mode-1 mode-2)))
+    (let ((f (namestring (merge-pathnames "some.file" *test-directory*))))
+      (close (open f :if-exists :append :if-does-not-exist :create))
+      (let* ((stat-1 (sb-posix:stat "/"))
+             (mode-1 (sb-posix:stat-mode stat-1))
+             (stat-2 (sb-posix:stat f stat-1))
+             (mode-2 (sb-posix:stat-mode stat-2)))
+        (values
+         (eq stat-1 stat-2)
+         (/= mode-1 mode-2))))
   t
   t)
 
@@ -331,11 +317,13 @@
             (,mode (sb-posix::stat-mode ,stat)))
        ,@body)))
 
+#-(and darwin x86)
 (deftest stat-mode.1
   (with-stat-mode (mode *test-directory*)
     (sb-posix:s-isreg mode))
   nil)
 
+#-(and darwin x86)
 (deftest stat-mode.2
   (with-stat-mode (mode *test-directory*)
     (sb-posix:s-isdir mode))
@@ -346,6 +334,7 @@
     (sb-posix:s-ischr mode))
   nil)
 
+#-(and darwin x86)
 (deftest stat-mode.4
   (with-stat-mode (mode *test-directory*)
     (sb-posix:s-isblk mode))
@@ -356,13 +345,13 @@
     (sb-posix:s-isfifo mode))
   nil)
 
-#-win32
+#-(or (and darwin x86) win32)
 (deftest stat-mode.6
   (with-stat-mode (mode *test-directory*)
     (sb-posix:s-issock mode))
   nil)
 
-#-win32
+#-(or (and darwin x86) win32)
 (deftest stat-mode.7
   (let ((link-pathname (make-pathname :name "stat-mode.7"
                                       :defaults *test-directory*)))
@@ -374,6 +363,7 @@
       (ignore-errors (sb-posix:unlink link-pathname))))
   t)
 
+#-(and darwin x86)
 (deftest stat-mode.8
   (let ((pathname (make-pathname :name "stat-mode.8"
                                  :defaults *test-directory*)))
@@ -391,6 +381,7 @@
   (let ((file (format nil "~A/[foo].txt" (namestring *test-directory*))))
     ;; creat() with a string as argument
     (let ((fd (sb-posix:creat file sb-posix:s-iwrite)))
+      (declare (ignorable fd))
       #+win32
       (sb-posix:close fd))
     ;; if this test fails, it will probably be with
@@ -411,7 +402,6 @@
         (ignore-errors (sb-posix:unlink name))))
   nil)
 
-#-hpux ; fix: cant handle c-vargs
 (deftest open.error.1
   (handler-case (sb-posix:open *test-directory* sb-posix::o-wronly)
     (sb-posix:syscall-error (c)
@@ -421,28 +411,23 @@
   #+win32
   #.sb-posix:eacces)
 
-#-(or (and x86-64 (or linux sunos)) win32)
-(deftest fcntl.1
-  (let ((fd (sb-posix:open "/dev/null" sb-posix::o-nonblock)))
-    (= (sb-posix:fcntl fd sb-posix::f-getfl) sb-posix::o-nonblock))
-  t)
-;; On AMD64/Linux O_LARGEFILE is always set, even though the whole
+;; O_LARGEFILE is always set on 64-bit *nix platforms even though the whole
 ;; flag makes no sense.
-#+(and x86-64 (or linux sunos))
+#-(or (and darwin x86) win32)
 (deftest fcntl.1
-  (let ((fd (sb-posix:open "/dev/null" sb-posix::o-nonblock)))
-    (/= 0 (logand (sb-posix:fcntl fd sb-posix::f-getfl)
-                  sb-posix::o-nonblock)))
+    (let ((fd (sb-posix:open "/dev/null" sb-posix::o-nonblock)))
+      (logtest (sb-posix:fcntl fd sb-posix::f-getfl)
+               sb-posix::o-nonblock))
   t)
 
-#-(or hpux win32) ; fix: cant handle c-vargs
+#-(or gc-stress win32 netbsd haiku) ; fix: cant handle c-vargs
 (deftest fcntl.flock.1
     (locally (declare (sb-ext:muffle-conditions sb-ext:compiler-note))
       (let ((flock (make-instance 'sb-posix:flock
                       :type sb-posix:f-wrlck
                       :whence sb-posix:seek-set
                       :start 0 :len 10))
-            (pathname "fcntl.flock.1")
+            (pathname (merge-pathnames #P"fcntl.flock.1" *test-directory*))
             kid-status)
         (catch 'test
           (with-open-file (f pathname :direction :output)
@@ -453,13 +438,13 @@
                   (progn
                     (multiple-value-bind (nope error)
                         (ignore-errors (sb-posix:fcntl f sb-posix:f-setlk flock))
-                      (sb-ext:quit
-                       :unix-status
+                      (sb-ext:exit
+                       :code
                        (cond ((not (null nope)) 1)
                              ((= (sb-posix:syscall-errno error) sb-posix:eagain)
                               42)
                              (t 86))
-                       :recklessly-p t #| don't delete the file |#)))
+                       :abort t #| don't delete the file |#)))
                   (progn
                     (setf kid-status
                           (sb-posix:wexitstatus
@@ -470,14 +455,14 @@
   42)
 
 
-#-win32
+#-(or gc-stress win32 netbsd)
 (deftest fcntl.flock.2
     (locally (declare (sb-ext:muffle-conditions sb-ext:compiler-note))
       (let ((flock (make-instance 'sb-posix:flock
                       :type sb-posix:f-wrlck
                       :whence sb-posix:seek-set
                       :start 0 :len 10))
-            (pathname "fcntl.flock.2")
+            (pathname (merge-pathnames #P"fcntl.flock.2" *test-directory*))
             kid-status)
         (catch 'test
           (with-open-file (f pathname :direction :output)
@@ -487,12 +472,12 @@
                   (pid (sb-posix:fork)))
               (if (zerop pid)
                   (let ((r (sb-posix:fcntl f sb-posix:f-getlk flock)))
-                    (sb-ext:quit
-                     :unix-status
+                    (sb-ext:exit
+                     :code
                      (cond ((not (zerop r)) 1)
                            ((= (sb-posix:flock-pid flock) ppid) 42)
                            (t 86))
-                     :recklessly-p t #| don't delete the file |#))
+                     :abort t #| don't delete the file |#))
                   (progn
                     (setf kid-status
                           (sb-posix:wexitstatus
@@ -502,6 +487,24 @@
         kid-status))
   42)
 
+(deftest read.1
+    (progn
+      (with-open-file (ouf (merge-pathnames "read-test.txt" *test-directory*)
+                           :direction :output
+                           :if-exists :supersede
+                           :if-does-not-exist :create)
+        (write-string "foo" ouf))
+      (let ((fd (sb-posix:open (merge-pathnames "read-test.txt" *test-directory*) sb-posix:o-rdonly)))
+        (unwind-protect
+             (let ((buf (make-array 10 :element-type '(unsigned-byte 8))))
+               (values
+                (sb-posix:read fd (sb-sys:vector-sap buf) 10)
+                (code-char (aref buf 0))
+                (code-char (aref buf 1))
+                (code-char (aref buf 2))))
+          (sb-posix:close fd))))
+  3 #\f #\o #\o)
+
 (deftest opendir.1
   (let ((dir (sb-posix:opendir "/")))
     (unwind-protect (sb-alien:null-alien dir)
@@ -509,6 +512,7 @@
         (sb-posix:closedir dir))))
   nil)
 
+#-(or (and darwin x86) haiku)
 (deftest readdir.1
   (let ((dir (sb-posix:opendir "/")))
     (unwind-protect
@@ -521,68 +525,158 @@
       (sb-posix:closedir dir)))
   t)
 
-#-darwin
-(deftest readdir/dirent-name
-    (let ((dir (sb-posix:opendir *current-directory*)))
-      (unwind-protect
-           (equal (sort (loop for entry = (sb-posix:readdir dir)
-                           until (sb-alien:null-alien entry)
-                           collect (sb-posix:dirent-name entry))
-                        #'string<)
-                  (sort (append '("." "..")
-                                (mapcar (lambda (p)
-                                          (let ((string (enough-namestring p *current-directory*)))
-                                            (if (pathname-name p)
-                                                string
-                                                (subseq string 0 (1- (length string))))))
-                                        (directory (make-pathname
-                                                    :name :wild
+#-(or darwin haiku)
+(test-util:with-test (:name :readdir/dirent-name)
+  (let* ((dir (sb-posix:opendir *current-directory*))
+         (posix-readdir (loop for entry = (sb-posix:readdir dir)
+                              until (sb-alien:null-alien entry)
+                              collect (sb-posix:dirent-name entry)))
+         (cl-directory
+          (append '("." "..")
+                  (mapcar (lambda (p)
+                            (let ((string (enough-namestring p *current-directory*)))
+                              (if (pathname-name p)
+                                  string
+                                  (subseq string 0 (1- (length string))))))
+                          (directory (make-pathname :name :wild
                                                     :type :wild
-                                                    :defaults *current-directory*))))
-                        #'string<))
-        (sb-posix:closedir dir)))
-  t)
+                                                    :defaults *current-directory*)
+                                     :resolve-symlinks nil)))))
+    (sb-posix:closedir dir)
+    (assert (equal (sort posix-readdir #'string<)
+                   (sort cl-directory #'string<)))))
 
-#-win32
+(test-util:with-test (:name :write.1)
+  (multiple-value-bind (n str)
+     (test-util:with-scratch-file (tmpname)
+      (let ((fd (sb-posix:open tmpname
+                               (logior sb-posix:o-creat sb-posix:o-wronly)
+                               (logior sb-posix:s-irusr sb-posix:s-iwusr)))
+            (retval nil))
+        (unwind-protect
+             (let ((buf (coerce "foo" 'simple-base-string)))
+               (setf retval (sb-posix:write fd (sb-sys:vector-sap buf) 3)))
+          (sb-posix:close fd))
+
+        (with-open-file (inf tmpname) (values retval (read-line inf)))))
+    (assert (and (eql n 3) (equal str "foo")))))
+
+#-(or android win32)
 (deftest pwent.1
   ;; make sure that we found something
   (not (sb-posix:getpwuid 0))
   nil)
 
-#-win32
+#-(or android win32 haiku)
 (deftest pwent.2
   ;; make sure that we found something
   (not (sb-posix:getpwnam "root"))
   nil)
 
-#-win32
+#-(or android win32)
 (deftest pwent.non-existing
     ;; make sure that we get something sensible, not an error
     (handler-case (progn (sb-posix:getpwnam "almost-certainly-does-not-exist")
                          nil)
-      (t (cond) t))
+      (t (cond) (declare (ignore cond)) t))
   nil)
 
-#-win32
+#-(or android win32 (not sb-thread))
+(deftest do-passwds.concurrency
+    (let* (thread1 thread2
+           (t1
+            (sb-posix:do-passwds (passwd)
+              ;; Both the two following threads should wait
+              ;; for WITH-PASSWD-DATABASE to exit.
+              (setq
+               thread1
+               (sb-thread:make-thread
+                (lambda ()
+                  (sb-posix:getpwnam "root")
+                  (get-universal-time)))
+               thread2
+               (sb-thread:make-thread
+                (lambda ()
+                  (sb-posix:getpwuid 0)
+                  (get-universal-time))))
+             (let ((ut (get-universal-time)))
+               (sleep 1.1)
+               (return ut))))
+           (t2 (sb-thread:join-thread thread1))
+           (t3 (sb-thread:join-thread thread2)))
+      (values (> t2 t1) (> t3 t1)))
+  t t)
+
+#-(or android win32)
+(deftest do-passwds.1
+    ;; Just check that we get something back.
+    (typep
+     (sb-posix:do-passwds (passwd)
+      (return passwd))
+     'sb-posix:passwd)
+  t)
+
+#-(or android win32)
 (deftest grent.1
   ;; make sure that we found something
   (not (sb-posix:getgrgid 0))
   nil)
 
-#-win32
+#-(or android win32)
 (deftest grent.2
-  ;; make sure that we found something
-  (not (sb-posix:getgrnam "sys"))
-  nil)
+    ;; make sure that we found something
+    (let* ((gid 0)
+           (group (sb-posix:getgrgid gid)))
+      (eql gid
+           (sb-posix:group-gid (sb-posix:getgrnam (sb-posix:group-name group)))))
+  t)
 
-#-win32
+#-(or android win32)
 (deftest grent.non-existing
     ;; make sure that we get something sensible, not an error
     (handler-case (progn (sb-posix:getgrnam "almost-certainly-does-not-exist")
                          nil)
-      (t (cond) t))
+      (t (cond) (declare (ignore cond)) t))
   nil)
 
+#-(or android win32 (not sb-thread))
+(deftest do-group-database/getgrnam/getgruid.concurrency
+    (let* (thread1 thread2
+           (group-name (let ((group (sb-posix:getgrgid 0)))
+                         (assert group) ;; see test grent.1
+                         (sb-posix:group-name group)))
+           (t1
+            (sb-posix:do-groups (group)
+              ;; Both the two following threads should wait
+              ;; for WITH-PASSWD-DATABASE to exit.
+              (setq
+               thread1
+               (sb-thread:make-thread
+                (lambda ()
+                  (sb-posix:getgrnam group-name)
+                  (get-universal-time)))
+               thread2
+               (sb-thread:make-thread
+                (lambda ()
+                  (sb-posix:getgrgid 0)
+                  (get-universal-time))))
+              (let ((ut (get-universal-time)))
+                (sleep 1.1)
+                (return ut))))
+           (t2 (sb-thread:join-thread thread1))
+           (t3 (sb-thread:join-thread thread2)))
+      (values (> t2 t1) (> t3 t1)))
+  t t)
+
+#-(or android win32)
+(deftest do-groups.1
+    ;; Just check that we get something back.
+    (typep
+     (sb-posix:do-groups (group)
+      (return group))
+     'sb-posix:group)
+  t)
+
 #+nil
 ;; Requires root or special group + plus a sensible thing on the port
 (deftest cfget/setispeed.1
@@ -619,22 +713,25 @@
     (plusp (sb-posix:time))
   t)
 
-#-win32
-(deftest utimes.1
-    (let ((file (merge-pathnames #p"utimes.1" *test-directory*))
-          (atime (random (1- (expt 2 31))))
-          (mtime (random (1- (expt 2 31)))))
-      (with-open-file (stream file
-                       :direction :output
-                       :if-exists :supersede
-                       :if-does-not-exist :create)
-        (princ "Hello, utimes" stream))
-      (sb-posix:utime file atime mtime)
-      (let* ((stat (sb-posix:stat file)))
-        (delete-file file)
-        (list (= (sb-posix:stat-atime stat) atime)
-              (= (sb-posix:stat-mtime stat) mtime))))
-  (t t))
+#-(or (and darwin x86) win32 haiku)
+(macrolet ((test (name posix-fun)
+             `(deftest ,name
+                (let ((file (merge-pathnames #p"utimes.1" *test-directory*))
+                      (atime (random (1- (expt 2 31))))
+                      (mtime (random (1- (expt 2 31)))))
+                  (with-open-file (stream file
+                                          :direction :output
+                                          :if-exists :supersede
+                                          :if-does-not-exist :create)
+                                  (princ "Hello, utimes" stream))
+                  (,posix-fun file atime mtime)
+                  (let* ((stat (sb-posix:stat file)))
+                    (delete-file file)
+                    (list (= (sb-posix:stat-atime stat) atime)
+                          (= (sb-posix:stat-mtime stat) mtime))))
+                (t t))))
+  (test utime.1 sb-posix:utime)
+  (test utimes.1 sb-posix:utimes))
 
 ;; readlink tests.
 #-win32
@@ -651,7 +748,6 @@
   ;; Same thing, but with a very long link target (which doesn't have
   ;; to exist).  This tests the array adjustment in the wrapper,
   ;; provided that the target's length is long enough.
-  #-hpux ; arg2 to readlink is 80, and arg0 is larger than that
   (deftest readlink.2
       (let ((target-pathname (make-pathname
                               :name (make-string 255 :initial-element #\a)
@@ -665,29 +761,33 @@
     #.(concatenate 'string "/" (make-string 255 :initial-element #\a)))
 
   ;; The error tests are in the order of exposition from SUSv3.
+  #-(or freebsd haiku)
   (deftest readlink.error.1
-      (let* ((subdir-pathname (merge-pathnames
-                               (make-pathname
-                                :directory '(:relative "readlink.error.1"))
-                               *test-directory*))
-             (link-pathname (make-pathname :name "readlink.error.1"
-                                           :defaults subdir-pathname)))
-        (sb-posix:mkdir subdir-pathname #o777)
-        (sb-posix:symlink "/" link-pathname)
-        (sb-posix:chmod subdir-pathname 0)
-        (unwind-protect
-             (handler-case (sb-posix:readlink link-pathname)
-               (sb-posix:syscall-error (c)
-                 (sb-posix:syscall-errno c)))
-          (ignore-errors
-            (sb-posix:chmod subdir-pathname #o777)
-            (sb-posix:unlink link-pathname)
-            (sb-posix:rmdir subdir-pathname))))
+    (if (zerop (sb-posix:getuid))
+        sb-posix:eacces
+        (let* ((subdir-pathname (merge-pathnames
+                                 (make-pathname
+                                  :directory '(:relative "readlink.error.1"))
+                                 *test-directory*))
+               (link-pathname (make-pathname :name "readlink.error.1"
+                                             :defaults subdir-pathname)))
+          (sb-posix:mkdir subdir-pathname #o777)
+          (sb-posix:symlink "/" link-pathname)
+          (sb-posix:chmod subdir-pathname 0)
+          (unwind-protect
+               (handler-case (sb-posix:readlink link-pathname)
+                 (sb-posix:syscall-error (c)
+                   (sb-posix:syscall-errno c)))
+            (ignore-errors
+             (sb-posix:chmod subdir-pathname #o777)
+             (sb-posix:unlink link-pathname)
+             (sb-posix:rmdir subdir-pathname)))))
     #.sb-posix:eacces)
+  #-haiku
   (deftest readlink.error.2
       (let* ((non-link-pathname (make-pathname :name "readlink.error.2"
                                                :defaults *test-directory*))
-             (fd (sb-posix:open non-link-pathname sb-posix::o-creat)))
+             (fd (sb-posix:open non-link-pathname sb-posix:o-creat #o777)))
         (unwind-protect
              (handler-case (sb-posix:readlink non-link-pathname)
                (sb-posix:syscall-error (c)
@@ -714,6 +814,7 @@
     #.sb-posix:eloop)
   ;; Note: PATH_MAX and NAME_MAX need not be defined, and may vary, so
   ;; failure of this test is not too meaningful.
+  #-haiku
   (deftest readlink.error.4
       (let ((pathname
              (make-pathname :name (make-string 257 ;NAME_MAX plus some, maybe
@@ -745,7 +846,7 @@
                                     '(:relative "readlink.error.7")
                                     :name "readlink.error.7")
                                    *test-directory*))
-             (fd (sb-posix:open non-link-pathname sb-posix::o-creat)))
+             (fd (sb-posix:open non-link-pathname sb-posix:o-creat #o777)))
         (unwind-protect
              (handler-case (sb-posix:readlink impossible-pathname)
                (sb-posix:syscall-error (c)
@@ -761,6 +862,22 @@
     (equal (sb-unix::posix-getcwd) (sb-posix:getcwd))
   t)
 
+(defun parse-native-namestring (namestring &key as-directory)
+  ;; XXXXXX can contain a dot changing pathname-type
+  (let ((parsed (sb-ext:parse-native-namestring namestring nil
+                                                *default-pathname-defaults*
+                                                :as-directory as-directory)))
+    (if as-directory
+        parsed
+        (let* ((name (pathname-name parsed))
+               (type (pathname-type parsed))
+               (dot (position #\. name)))
+          (if dot
+              (make-pathname :name (subseq name 0 dot)
+                             :type (format nil "~a.~a" (subseq name (1+ dot)) type)
+                             :defaults parsed)
+              parsed)))))
+
 #-win32
 (deftest mkstemp.1
     (multiple-value-bind (fd temp)
@@ -768,25 +885,22 @@
                            :name "mkstemp-1"
                            :type "XXXXXX"
                            :defaults *test-directory*))
-      (let ((pathname (sb-ext:parse-native-namestring temp)))
+      (let ((pathname (parse-native-namestring temp)))
         (unwind-protect
              (values (integerp fd) (pathname-name pathname))
           (delete-file temp))))
   t "mkstemp-1")
 
-;#-(or win32 sunos hpux)
 ;;;; mkdtemp is unimplemented on at least Solaris 10
-#-(or win32 hpux sunos)
+#-(or win32 sunos)
 ;;; But it is implemented on OpenSolaris 2008.11
 (deftest mkdtemp.1
     (let ((pathname
-           (sb-ext:parse-native-namestring
+           (parse-native-namestring
             (sb-posix:mkdtemp (make-pathname
                                :name "mkdtemp-1"
                                :type "XXXXXX"
                                :defaults *test-directory*))
-            nil
-            *default-pathname-defaults*
             :as-directory t)))
       (unwind-protect
            (values (let* ((xxx (car (last (pathname-directory pathname))))
@@ -799,7 +913,7 @@
 
 #-win32
 (deftest mktemp.1
-    (let ((pathname (sb-ext:parse-native-namestring
+    (let ((pathname (parse-native-namestring
                      (sb-posix:mktemp #p"mktemp.XXXXXX"))))
       (values (equal "mktemp" (pathname-name pathname))
               (not (equal "XXXXXX" (pathname-type pathname)))))
@@ -821,11 +935,10 @@
                              :type (format nil "~AXXXXXX"
                                            (make-string n :initial-element #\x))
                              :defaults default))
-        (let ((pathname (sb-ext:parse-native-namestring temp)))
           (unwind-protect
-               (values (integerp fd) (pathname-name pathname))
-            (delete-file temp)))))
-  t "mkstemp-1")
+               (values (integerp fd) (subseq temp 0 (position #\. temp)))
+            (delete-file temp))))
+  t "/tmp/mkstemp-1")
 
 (deftest envstuff
     (let ((name1 "ASLIFJLSDKFJKAHGSDKLJH")
