@@ -55,7 +55,17 @@
 #+arm-softfp
 (define-alien-type-method (single-float :arg-tn) (type state)
   (declare (ignore type))
-  (int-arg state 'single-float unsigned-reg-sc-number single-stack-sc-number))
+  (let ((register (arg-state-num-register-args state)))
+    (cond ((>= register +max-register-args+)
+           (let ((frame-size (arg-state-stack-frame-size state)))
+             (incf (arg-state-stack-frame-size state))
+             (make-wired-tn* 'single-float single-stack-sc-number frame-size)))
+          (t
+           (incf (arg-state-num-register-args state))
+           (list
+            (make-wired-tn* 'unsigned-byte-32 unsigned-reg-sc-number
+                            (register-args-offset register))
+            'move-single-to-int-args)))))
 
 #-arm-softfp
 (define-alien-type-method (single-float :arg-tn) (type state)
@@ -123,7 +133,9 @@
 #+arm-softfp
 (define-alien-type-method (single-float :result-tn) (type state)
   (declare (ignore type state))
-  (make-wired-tn* 'single-float unsigned-reg-sc-number nargs-offset))
+  (list (make-wired-tn* 'unsigned-byte-32 unsigned-reg-sc-number nargs-offset)
+        (make-normal-tn (primitive-type-or-lose 'single-float))
+        'move-int-args-to-single))
 
 #-arm-softfp
 (define-alien-type-method (single-float :result-tn) (type state)
@@ -247,26 +259,46 @@
 ;;;
 
 #+arm-softfp
-(define-vop (move-double-to-int-args)
-  (:args (double :scs (double-reg)))
-  (:results (lo-bits :scs (unsigned-reg))
-            (hi-bits :scs (unsigned-reg)))
-  (:arg-types double-float)
-  (:result-types unsigned-num unsigned-num)
-  (:policy :fast-safe)
-  (:generator 1
-    (inst fmrrd lo-bits hi-bits double)))
+(progn
+  (define-vop (move-double-to-int-args)
+    (:args (double :scs (double-reg)))
+    (:results (lo-bits :scs (unsigned-reg))
+              (hi-bits :scs (unsigned-reg)))
+    (:arg-types double-float)
+    (:result-types unsigned-num unsigned-num)
+    (:policy :fast-safe)
+    (:generator 1
+      (inst fmrrd lo-bits hi-bits double)))
 
-#+arm-softfp
-(define-vop (move-int-args-to-double)
-  (:args (lo-bits :scs (unsigned-reg))
-         (hi-bits :scs (unsigned-reg)))
-  (:results (double :scs (double-reg)))
-  (:arg-types unsigned-num unsigned-num)
-  (:result-types double-float)
-  (:policy :fast-safe)
-  (:generator 1
-    (inst fmdrr double lo-bits hi-bits)))
+
+  (define-vop (move-int-args-to-double)
+    (:args (lo-bits :scs (unsigned-reg))
+           (hi-bits :scs (unsigned-reg)))
+    (:results (double :scs (double-reg)))
+    (:arg-types unsigned-num unsigned-num)
+    (:result-types double-float)
+    (:policy :fast-safe)
+    (:generator 1
+      (inst fmdrr double lo-bits hi-bits)))
+
+  (define-vop (move-single-to-int-args)
+    (:args (single :scs (single-reg)))
+    (:results (bits :scs (unsigned-reg)))
+    (:arg-types single-float)
+    (:result-types unsigned-num)
+    (:policy :fast-safe)
+    (:generator 1
+      (inst fmrs bits single)))
+
+
+  (define-vop (move-int-args-to-single)
+    (:args (bits :scs (unsigned-reg)))
+    (:results (single :scs (single-reg)))
+    (:arg-types unsigned-num)
+    (:result-types single-float)
+    (:policy :fast-safe)
+    (:generator 1
+      (inst fmsr single bits))))
 
 ;;; long-long support
 (deftransform %alien-funcall ((function type &rest args) * * :node node)
@@ -343,25 +375,12 @@
 ;;; Callback
 #-sb-xc-host
 (defun alien-callback-accessor-form (type sap offset)
-  (let ((parsed-type type))
-    (if (alien-integer-type-p parsed-type)
-        (let ((bits (sb-alien::alien-integer-type-bits parsed-type)))
-               (let ((byte-offset
-                      (cond ((< bits n-word-bits)
-                             (- n-word-bytes
-                                (ceiling bits n-byte-bits)))
-                            (t 0))))
-                 `(deref (sap-alien (sap+ ,sap
-                                          ,(+ byte-offset offset))
-                                    (* ,type)))))
-        `(deref (sap-alien (sap+ ,sap ,offset) (* ,type))))))
+  `(deref (sap-alien (sap+ ,sap ,offset) (* ,type))))
 
 #-sb-xc-host
 (defun alien-callback-assembler-wrapper (index result-type argument-types)
   (flet ((make-tn (offset &optional (sc-name 'any-reg))
-           (make-random-tn :kind :normal
-                           :sc (sc-or-lose sc-name)
-                           :offset offset)))
+           (make-random-tn (sc-or-lose sc-name) offset)))
     (let* ((segment (make-segment))
            ;; How many arguments have been copied
            (arg-count 0)
